@@ -15,18 +15,19 @@
 #
 #   Libraries:      None (only core)
 #
-#   Global Consts:  $TOKEN_RE
+#   Global Consts:  @EXPORT_OK
+#                   %NOT_ACCEPTABLE_REF
 #
 #   Environment:    None
 #
 ###############################################################################
 
-use 5.014;
 use strict;
 use warnings;
 
 package YASF;
-$YASF::VERSION = '0.002';
+$YASF::VERSION = '0.003';
+use 5.008;
 use overload fallback => 0,
     'eq'  => \&_eq,
     'ne'  => \&_ne,
@@ -50,8 +51,19 @@ BEGIN {
     }
 }
 
+my %NOT_ACCEPTABLE_REF = (
+    SCALAR  => 1,
+    CODE    => 1,
+    REF     => 1,
+    GLOB    => 1,
+    LVALUE  => 1,
+    FORMAT  => 1,
+    IO      => 1,
+    VSTRING => 1,
+    Regexp  => 1,
+);
+
 our @EXPORT_OK = qw(YASF);
-my $TOKEN_RE = qr/(?^:((?:(?<!\\)[{](?:(?>[^{}]+)|(?-1))*(?<!\\)[}])))/x;
 
 ###############################################################################
 #
@@ -113,6 +125,8 @@ sub new {
 #                   $self     in      ref       Object of this class
 #                   $bindings in      ref       New bindings
 #
+#   Globals:        %NOT_ACCEPTABLE_REF
+#
 #   Returns:        Success:    $self
 #                   Failure:    dies
 #
@@ -120,18 +134,6 @@ sub new {
 sub bind { ## no critic(ProhibitBuiltinHomonyms)
     my ($self, $bindings) = @_;
 
-    state $not_acceptable = {
-        SCALAR  => 1,
-        CODE    => 1,
-        REF     => 1,
-        GLOB    => 1,
-        LVALUE  => 1,
-        FORMAT  => 1,
-        IO      => 1,
-        VSTRING => 1,
-        Regexp  => 1,
-    };
-
     if ((@_ == 2) && (! defined $bindings)) {
         # The means of unbinding is to call $obj->bind(undef):
         undef $self->{binding};
@@ -142,7 +144,7 @@ sub bind { ## no critic(ProhibitBuiltinHomonyms)
         my $type = ref $bindings;
         if (! $type) {
             croak 'New bindings must be a reference (HASH, ARRAY or object)';
-        } elsif ($not_acceptable->{$type}) {
+        } elsif ($NOT_ACCEPTABLE_REF{$type}) {
             croak "New bindings reference type ($type) not usable";
         }
 
@@ -173,14 +175,12 @@ sub bind { ## no critic(ProhibitBuiltinHomonyms)
 sub format { ## no critic(ProhibitBuiltinHomonyms)
     my ($self, $bindings) = @_;
 
-    ## no critic(BuiltinFunctions::ProhibitUselessTopic)
-
     $bindings ||= $self->binding;
     croak 'format: Bindings are required if object has no internal binding'
         if (! $bindings);
 
     my $value = join q{} =>
-        map { ref($_) ? $self->_format($bindings, @{$_}) : $_ }
+        map { ref() ? $self->_format($bindings, @{$_}) : $_ }
         @{$self->{_compiled}};
 
     return $value;
@@ -188,53 +188,84 @@ sub format { ## no critic(ProhibitBuiltinHomonyms)
 
 # Private functions that support the public API:
 
-# Front-end to the recursive _compile_segment:
+# Compile the template into a tree structure that can be easily traversed for
+# formatting.
 sub _compile {
     my $self = shift;
 
-    $self->{_compiled} = $self->_compile_segment($self->template);
+    my @tokens = $self->_tokenize;
 
-    return;
+    my @stack = ([]);
+    my $level = 0;
+    my @opens = ();
+
+    while (my ($type, $value) = splice @tokens, 0, 2) {
+        if ($type eq 'STRING') {
+            push @{$stack[$level]}, $value;
+        } elsif ($type eq 'OPEN') {
+            push @stack, [];
+            $level++;
+            push @opens, $value;
+        } elsif ($type eq 'CLOSE') {
+            if ($level) {
+                my $subtree = pop @stack;
+                $level--;
+                push @{$stack[$level]}, $subtree;
+                pop @opens;
+            } else {
+                croak "Unmatched closing brace at position $value";
+            }
+        } else {
+            croak "Unknown token type: $type\n";
+        }
+    }
+
+    if ($level) {
+        croak sprintf '%d unmatched opening brace%s, last at position %d',
+            $level, $level == 1 ? q{} : 's', pop @opens;
+    }
+
+    $self->{_compiled} = $stack[0];
+    return $self;
 }
 
-# Compiles a segment of the template. Creates an arrayref of constant parts
-# (strings) and nested expansion parts (arrayrefs). Uses the $TOKEN_RE global.
-sub _compile_segment {
-    my ($self, $segment) = @_;
-    my (@tokens, @compiled, $pos);
+# Tokenize the object's template into a sequence of (type, value) pairs that
+# identify the opening and closing braces, and ordinary strings.
+sub _tokenize {
+    my $self = shift;
 
-    while ($segment =~ /$TOKEN_RE/g) {
-        push @tokens, [ $1, $LAST_MATCH_START[1], $LAST_MATCH_END[1] ];
-    }
+    my (@list, $base, $pos, $len);
+    my $str = $self->template;
 
-    $pos = 0;
-    for my $token (@tokens) {
-        my ($subsegment, $start, $end) = @{$token};
-        if (my $len = $start - $pos) {
-            push @compiled, substr $segment, $pos, $len;
+    $base = 0;
+    while ($str =~ /(?<!\\)(?=[{}])/g) {
+        $pos = pos $str;
+        if ($len = $pos - $base) {
+            (my $piece = substr $str, $base, $len) =~ s/\\([{}])/$1/g;
+            push @list, 'STRING', $piece;
         }
-        push @compiled, $self->_compile_segment(substr $subsegment, 1, -1);
-        $pos = $end;
+        push @list, ('{' eq substr $str, $pos, 1) ? 'OPEN' : 'CLOSE';
+        push @list, $pos;
+        $base = $pos + 1;
     }
 
-    if ($pos < length $segment) {
-        push @compiled, substr $segment, $pos;
+    if (length($str) > $base) {
+        (my $piece = substr $str, $base) =~ s/\\([{}])/$1/g;
+        push @list, 'STRING', $piece;
     }
 
-    return \@compiled;
+    return @list;
 }
 
 # Does the hard and recursive part of the actual formatting. Not actually that
-# hard, but pretty recursive.
+# hard, but a little recursive.
 sub _format {
     my ($self, $bindings, @elements) = @_;
 
-    ## no critic(BuiltinFunctions::ProhibitUselessTopic)
-
     # Slight duplication of code from format() here, but it saves having to
     # keep track of depth and do a conditional on every return.
     my $expr = join q{} =>
-        map { ref($_) ? $self->_format($bindings, @{$_}) : $_ } @elements;
+        map { ref() ? $self->_format($bindings, @{$_}) : $_ } @elements;
 
     return $self->_expr_to_value($bindings, $expr);
 }
@@ -371,7 +402,7 @@ YASF - Yet Another String Formatter
 
 =head1 VERSION
 
-version 0.002
+version 0.003
 
 =head1 SYNOPSIS
 
@@ -441,6 +472,24 @@ endless recursion. That is, if C<subkey> evaluated to C<{key2}>, it would
 B<not> result in C<key2> being interpolated. Instead, a literal key of
 C<{key2}> would be looked up on the hash reference that C<key> yields.
 
+Because curly braces are used to delimit keys, if you want a curly brace (of
+either orientation) in your string you will need to escape it with a
+backslash character:
+
+    $str = YASF->new('\{{key}\}');
+
+or
+
+    $str = YASF->new("\\{{key}\\}");
+
+The escaped characters will be converted when the template is compiled
+internally.
+
+A key may be made up of any characters, though for readability it is
+recommended that you stay with alphanumerics. The only character that cannot
+be used in a key is the colon (C<:>), as this is used to delimit a key from
+a formatting specification (see L</Formatting Syntax>).
+
 =head2 Using Objects in the Bindings
 
 If an element within the bindings data structure is an object, the key for
@@ -468,11 +517,13 @@ binding:
 =head2 Formatting Syntax
 
 Python's C<format> also supports an extensive syntax for formatting the data
-that gets substituted.
+that gets substituted. A format specification is given in the key, as a string
+sequence separated from the key itself by a colon:
 
-This is not provided in this initial release of B<YASF>, but will be added
-in a future release. For now, if a formatting string is detected it will be
-ignored.
+    {key1.key2:6.2f}
+
+This is not yet implemented in B<YASF>, but will be added in a future release.
+For now, if a formatting string is detected it will be ignored.
 
 =head1 OVERLOADED OPERATORS
 
